/*******************************************************************************
 * Copyright (c) 2003, 2010 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

// @(#) $Id:  $
package at.spardat.xma.mdl.paging;

import java.io.IOException;
import java.util.List;

import at.spardat.xma.mdl.ModelChangeEvent;
import at.spardat.xma.mdl.NewModelEvent;
import at.spardat.xma.mdl.NewModelEventFactory;
import at.spardat.xma.mdl.WModel;
import at.spardat.xma.mdl.table.ITableWM;
import at.spardat.xma.mdl.table.TableWM;
import at.spardat.xma.mdl.util.TransNumber;
import at.spardat.xma.page.Page;
import at.spardat.xma.serializer.XmaInput;
import at.spardat.xma.serializer.XmaOutput;

/**
 * Implements the part of the paging model which is available on server and client side.
 * All this data is automatically synchronized between client and server. So you can let the user
 * modify this data using the GUI and query it on the server for loading the required slice of the
 * data objects to show in the corresponding table.
 *
 * @author gub
 * @since 2.3.0
 */
public class PagingWM extends WModel implements IPagingWM {
    /** Index of the first data object to show in the result list. */
    TransNumber offset = new TransNumber(TransNumber.T_INT,0);

    /** Number of data objects to show at the same time. */
    TransNumber pageSize = new TransNumber(TransNumber.T_SHORT,100);

    /** Maximum pageSize the user can choose. */
    TransNumber maxPageSize = new TransNumber(TransNumber.T_SHORT,1000);

    /** Total number of data objects in the result list. */
    TransNumber resultSize = new TransNumber(TransNumber.T_INT,Integer.MAX_VALUE);

    /** Number of pages to jump forward or back when the user presses the fastNext or fastBack button. */
    TransNumber jumpSize = new TransNumber(TransNumber.T_SHORT, 5);

    /** Index of the column to sort. */
    TransNumber sortingColumn = new TransNumber(TransNumber.T_SHORT,-1);

    /** Defines if sorting should be ascending or descending. */
    TransNumber ascending = new TransNumber(TransNumber.T_BYTE,1);

    /** Holds all members for synchronization. */
    TransNumber[] data;

    /** Flag if this was instantiated on the server side. */
    boolean isAtServer=true;

    /** Flag if the corresponding table was changed on server side. */
    boolean tableChanged;

    /** Corresponding table holding the paged data. */
    ITableWM table;

    /**
     * Constructs the paging model.
     * @param id a unique id identifying this WidgetModel within its page.
     * @param pm the {@link Page} this model belongs to.
     * @param table the corresponding table holding the data.
     */
    public PagingWM(short id, Page pm, ITableWM table) {
        super(id, pm);
        this.table=table;
        data = new TransNumber[] { offset, pageSize, maxPageSize, resultSize, jumpSize, sortingColumn, ascending };
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getOffset()
     */
    public int getOffset() {
        return offset.toInt();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setOffset(int)
     */
    public void setOffset(int offset) {
        this.offset.set(offset);
        handle(new PagingModelChangeEvent(this, false));
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getPageSize()
     */
    public short getPageSize() {
        return pageSize.toShort();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setPageSize(short)
     */
    public void setPageSize(short pageSize) {
        this.pageSize.set(pageSize);
        handle(new PagingModelChangeEvent(this, false));
    }

    /**
     * Used by UIDelegate to set the page size.
     * This method does not actualize the GUI.
     */
    void setPageSizeFromUI(short pageSize) {
        this.pageSize.set(pageSize);
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getMaxPageSize()
     */
    public short getMaxPageSize() {
        return maxPageSize.toShort();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setMaxPageSize(short)
     */
    public void setMaxPageSize(short maxPageSize) {
        this.maxPageSize.set(maxPageSize);
        if(maxPageSize<getPageSize()) {
            setPageSize(maxPageSize);
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getResultSize()
     */
    public int getResultSize() {
        return resultSize.toInt();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setResultSize(int)
     */
    public void setResultSize(int resultSize) {
        this.resultSize.set(resultSize);
        handle(new PagingModelChangeEvent(this, false));
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getJumpSize()
     */
    public short getJumpSize() {
        return jumpSize.toShort();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setJumpSize(short)
     */
    public void setJumpSize(short jumpSize) {
        this.jumpSize.set(jumpSize);
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getSortingColumn()
     */
    public short getSortingColumn() {
        return sortingColumn.toShort();
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setSortingColumn(short)
     */
    public void setSortingColumn(short sortingColumn) {
        this.sortingColumn.set(sortingColumn);
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getAscending()
     */
    public boolean getAscending() {
        return ascending.toByte()!=0;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#setAscending(boolean)
     */
    public void setAscending(boolean ascending) {
        this.ascending.set(ascending?1:0);
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.paging.IPagingWM#getTable()
     */
    public ITableWM getTable() {
        return table;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#changed()
     */
   public boolean changed() {
        for(int i=0;i<data.length;i++) {
            if(data[i].changed()) return true;
        }
        if(isAtServer) return ((TableWM)table).changed();
        else return false;
    }

   /* (non-Javadoc)
    * @see at.spardat.xma.mdl.WModel#clear()
    */
    public void clear() {
        // TODO Auto-generated method stub

    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#commit()
     */
    public void commit() {
        for(int i=0;i<data.length;i++) {
            data[i].commit();
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#equalsCS()
     */
    public void equalsCS(WModel mServer, int syncPoint) {
        // TODO Auto-generated method stub

    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#estimateMemory()
     */
    public int estimateMemory() {
        // 5*4 + 5*18 + 4
        return 114;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#handle()
     */
    public boolean handle(ModelChangeEvent event) {
        return true;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#randomlyChange()
     */
    public void randomlyChange() {
        // TODO Auto-generated method stub

    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.WModel#rollback()
     */
    public void rollback() {
        for(int i=0;i<data.length;i++) {
            data[i].rollback();
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.Synchronization#externalize(XmaOutput,boolean)
     */
    public void externalize(XmaOutput xo, boolean forceFull) throws IOException {
        if(forceFull) {
            xo.writeByte("num",data.length);
        } else {
            int numtoWrite=0;
            for(byte i=0;i<data.length;i++) {
                if(data[i].changed()) numtoWrite++;
            }
            xo.writeByte("num",numtoWrite);
        }
        for(byte i=0;i<data.length;i++) {
            if(forceFull||data[i].changed()) {
                xo.writeByte("index", i);
                data[i].externalize(xo, forceFull);
            }
        }
        if(isAtServer) {
            xo.writeBoolean("tableChanged",((TableWM)table).changed());
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.Synchronization#internalize(XmaInput)
     */
    public void internalize(XmaInput in, List formattable) throws IOException {
        byte numToRead = in.readByte();
        for(int i=0;i<numToRead;i++) {
            byte index = in.readByte();
            data[index].internalize(in, null);
        }
        if(!isAtServer) {
            tableChanged=in.readBoolean();
        }
        handle(new PagingModelChangeEvent(this,false));
    }

    /**
     * Generic Event for all changes in the paging model.
     * Used to notify the UI of changes.
     */
    public static class PagingModelChangeEvent extends ModelChangeEvent {
        /**
         * Creates a change event
         * @param destination the changed model
         * @param fromUI true if caused by UI, false otherwise
         */
        public PagingModelChangeEvent(WModel destination, boolean fromUI) {
            super(destination, fromUI);
        }
        /**
         * Does nothing since it is just a notification.
         */
        public boolean execute() {
            return true;
        }
    }

    /**
     * Event class used to notify the dynamic registration of a new PagingWM.
     * @see Page#addWModel(WModel)
    */
    public static class NewPagingWMEvent extends NewModelEvent {
        short tableId;

        /** empty contructor for deserialization */
        public NewPagingWMEvent() {}

        /**
         * constructor which initializes the table reference
         * @param tableId the model id of the corresponding table
         */
        public NewPagingWMEvent(short tableId) {
            this.tableId=tableId;
        }

        // see at.spardat.xma.mdl.NewModelEvent.getType()
        public byte getType() {
            return NewModelEventFactory.PagingWM;
        }

        // see at.spardat.xma.mdl.NewModelEvent.createModel()
        public WModel createModel(short id,Page page) {
            TableWM table = (TableWM) page.getWModel(tableId);
            return new PagingWM(id,page,table);
        }

        // see at.spardat.xma.mdl.NewModelEvent.serialize()
        public void serialize(XmaOutput out) throws IOException {
            super.serialize(out);
            out.writeShort("tableId", tableId);
        }

        // see at.spardat.xma.mdl.NewModelEvent.deserialize()
        public void deserialize(XmaInput in) throws IOException, ClassNotFoundException {
            super.deserialize(in);
            tableId=in.readShort();
        }
    }

    // see at.sparda.xma.mdl.WModel.createNewModelEvent()
    public NewModelEvent createNewModelEvent() {
        return new NewPagingWMEvent(((TableWM)table).getId());
    }

}
