/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

// @(#) $Id: Page.java 10268 2013-01-11 10:34:57Z dschwarz $
package at.spardat.xma.page;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import at.spardat.enterprise.exc.SysException;
import at.spardat.xma.mdl.IWModelClient;
import at.spardat.xma.mdl.NewModelEvent;
import at.spardat.xma.mdl.NewModelEventFactory;
import at.spardat.xma.mdl.NewModelEventParams;
import at.spardat.xma.mdl.Synchronization;
import at.spardat.xma.mdl.Transactional;
import at.spardat.xma.mdl.WModel;
import at.spardat.xma.mdl.util.DNode;
import at.spardat.xma.mdl.util.Descriptive;
import at.spardat.xma.serializer.XmaInput;
import at.spardat.xma.serializer.XmaOutput;
import at.spardat.xma.session.XMASession;


/**
 * Base class for server and client side PageModels. A Page is a set of
 * WidgetModels.
 */
public abstract class Page implements Transactional, Synchronization, Descriptive {

    /**
     * Constructs a Page.
     *
     * @param stateless indicates if this is a server stateless Page.
     * @param atServer indicates if the created instance is living at the server side.
     */
    public Page (boolean stateless, boolean atServer) {
        isStateless_ = stateless;
        isAtServer_ = atServer;
    }


    /**
     * Indicates whether this Page is stateless. This property specifies
     * that the state of the page is dropped at the server side after a
     * server side event.
     *
     * @return true if stateless
     */
    public boolean isStateless() {
        return isStateless_;
    }

    /**
     * Yields true if any of this PageModels WidgetModels has changed since the
     * last server side event (or since creation if isNew).
     *
     * @return true if changed.
     */
    public boolean changed() {
        if(newModelEvents_!=null&&newModelEvents_.size()>0) return true;
        WModel[]        wModels = getWModels();
        for (int i=wModels.length-1; i>=0; i--) if (wModels[i].changed()) return true;
        return false;
    }

    /**
     * Restores all WidgetModels of this PageModels to the state immediately
     * after the last server side event (or at creation time if isNew()).
     */
    public void rollback() {
        WModel[]        wModels = getWModels();
        for (int i=wModels.length-1; i>=0; i--) wModels[i].rollback();
    }

    /**
     * This method discards the change history in every WidgetModel.
     */
    public void commit() {
        WModel[]        wModels = getWModels();
        for (int i=wModels.length-1; i>=0; i--) wModels[i].commit();
        newModelEvents_=null;
    }

    /**
     * This method writes either all widget models if <tt>forceFull</tt> or just the changed ones.
     *
     * @see at.spardat.xma.mdl.Synchronization#externalize(at.spardat.xma.serializer.XmaOutput, boolean)
     */
    public void externalize (XmaOutput xo, boolean forceFull) throws IOException {
        List newModelList;
        if(forceFull) newModelList=handledNewModelEvents_;
        else newModelList=newModelEvents_;
        // write model creations
        if(newModelList==null) {
            xo.writeShort("numNewModels",0);
        } else {
            xo.writeShort("numNewModels",newModelList.size());
            for(Iterator it=newModelList.iterator();it.hasNext();) {
                NewModelEvent event = (NewModelEvent) it.next();
                xo.writeByte("type",event.getType());
                event.serialize(xo);
            }
        }
        // write models
        WModel[]            wModels = getWModels();
        int                 numToWrite = wModels.length;  // number of models if forceFull
        if (!forceFull) {
            numToWrite = 0;
            for (int i=0; i<wModels.length; i++)
                if (wModels[i].changed()) numToWrite++;
        }
        // write the number of widget models to externalize
        xo.writeShort("numModels", numToWrite);
        // write all models or just the changed
        for (int i=0; i<wModels.length; i++) {
            if (forceFull || wModels[i].changed()) {
                // write the model id
                xo.writeShort("id", i);
                wModels[i].externalize(xo, forceFull);
            }
        }
    }

    /**
     * @see at.spardat.xma.mdl.Synchronization#internalize(at.spardat.xma.serializer.XmaInput, List)
     */
    public void internalize (XmaInput in, List formattable) throws IOException, ClassNotFoundException {
        // read new models
        short numNewModels = in.readShort();
        if(numNewModels>0) {
            XMASession session = getSession();
            NewModelEventFactory eventFactory; // session is only null in unit tests
            if(session!=null) { eventFactory = getSession().getNewModelEventFactory(); }
            else { eventFactory = new NewModelEventFactory(); }
            for(int i=0;i<numNewModels;i++) {
                byte type = in.readByte();
                NewModelEvent event = eventFactory.createEvent(type,isAtServer_);
                event.deserialize(in);
                handle(event);
            }
        }
        // read models
        WModel[]            wModels = getWModels();
        // how many widget models must be read
        short               numToRead = in.readShort();
        for (int j=0; j<numToRead; j++) {
            // read the id of the widget model
            short           id = in.readShort();
            // read the widget model itself
            wModels[id].internalize(in, formattable);
        }
    }

    /**
     * returns the XMASession this page belongs to
     * @since 2.1.0
     */
    public abstract XMASession getSession();

    /**
     * Creates a new widget model corresponding to the given event and
     * integrates it into synchronization mechanism.
     * @param event sent from the other side
     * @return the new widget model
     * @since 2.1.0
     */
    WModel handle(NewModelEvent event) {
        WModel widgetModel = event.createModel(getNextModelId(), this);
        appendWModel(widgetModel);
        if(handledNewModelEvents_==null) handledNewModelEvents_=new ArrayList();
        handledNewModelEvents_.add(event);
        return widgetModel;
    }

    /**
     * Every page model has an integer id which identifies the page model
     * within its component.
     *
     * @return an id uniquely idenfying the page within its component.
     */
    public short getId() {
        return id_;
    }

    /**
     * Assigns an id to this Page. This method may only be called by
     * the Component.
     *
     * @param id the assigned id for this Page
     * @throws IllegalStateException if there is allready set an id.
     */
    public void setId(short id) {
        // if(id_!=0) throw new IllegalStateException("id allready set");
        id_ = id;
    }


    /**
     * Returns the number of Widget Models in this Page. The ids of the
     * WidgetModels range from 0 up to getNumWModels()-1.
     *
     * @return the number of WidgetModels this Page contains.
     */
    public int getNumWModels() {
        return getWModels().length;
    }

    /**
     * Returns a particular WidgetModel.
     *
     * @param idWModel the id of the WidgetModel to return.
     * @return the requested WidgetModel.
     * @exception IllegalArgumentException if <tt>idModel</tt> less then zero or
     *             greater than <tt>getNumWModels()-1</tt>.
     */
    public  WModel getWModel (short idWModel) {
        return getWModels()[idWModel];
    }

    /**
     * Returns an array of all WidgetModels of this page.
     * @return array of WidgetModels. The returned array is not a copy, therefore
     *          it must not be modified.
     */
    public abstract WModel[] getWModels();

    /**
     * Modifies the widget model array generated by guidesigner.
     * May only be called by the runtime. Do not call this method!
     * @since 2.1.0
     */
    protected void setWModels(WModel[] wModels) {
        throw new SysException("not supported");
    }

    /**
     * Appends a new widget model to WModel.
     * @since 2.1.0
     */
    private void appendWModel(WModel widgetModel) {
        WModel[] oldWModels = getWModels();
        if(widgetModel.getId()!=oldWModels.length) {
            throw new IllegalArgumentException("new widget model has invalid ID: "+widgetModel.getId());
        }
        WModel[] newWModels = new WModel[oldWModels.length+1];
        for(int i=0;i<oldWModels.length;i++) {
            newWModels[i]=oldWModels[i];
        }
        newWModels[oldWModels.length]=widgetModel;
        setWModels(newWModels);
    }

    /**
     * Adds a new widget model at runtime. A corresponding widget model is created automatically
     * on the other side and this new widget model on the other side is automatically synchronized
     * with the widgetModel passed as parameter.
     * On client side the method {@link PageClient#createWidgetForModel(IWModelClient, NewModelEventParams)}
     * is called to create a corresponding SWT widget, too.
     * @param widgetModel the new widget model which shall be integrated in automatically synchronizations.
     * @since 2.1.0
     */
    public void addWModel(WModel widgetModel) {
        addWModel(widgetModel,null);
    }

    /**
     * Adds a new widget model at runtime. A corresponding widget model is created automatically
     * on the other side and this new widget model on the other side is automatically synchronized
     * with the widgetModel passed as parameter.
     * On client side the method {@link PageClient#createWidgetForModel(IWModelClient, NewModelEventParams)}
     * is called to create a corresponding SWT widget, too.
     * @param widgetModel the new widget model which shall be integrated in automatically synchronizations.
     * @param params to be passed to {@link PageClient#createWidgetForModel(IWModelClient, NewModelEventParams)}
     * @since 2.1.0
     */
    public void addWModel(WModel widgetModel,NewModelEventParams params) {
        appendWModel(widgetModel);
        NewModelEvent event = widgetModel.createNewModelEvent();
        if(params!=null) event.setParams(params);
        if(newModelEvents_==null) newModelEvents_=new ArrayList();
        newModelEvents_.add(event);
        if(handledNewModelEvents_==null) handledNewModelEvents_=new ArrayList();
        handledNewModelEvents_.add(event);
    }

    /**
     * Returns the next free model id. When creating a widget model, to be added at runtime
     * using {@link #addWModel(WModel)}, this method must be used to determine the ID of the
     * new widget model.
     * @since 2.1.0
     */
    public short getNextModelId() {
        return (short) getWModels().length;
    }

    /**
     * A subclass is responsible for generating a unique id to identify the type
     * of this Page. With the knowledge of type id and component it is
     * possible to newly contruct a particular instance of a Page.
     *
     * @return the type id
     */
    public abstract short getTypeId ();


    /**
     * Indicates whether this Page executes at the server side.
     *
     * @return true, if this executes at the server side.
     */
    public boolean isAtServer() {
        return isAtServer_;
    }

    /**
     * @see at.spardat.xma.mdl.util.Descriptive#describe(at.spardat.xma.mdl.util.DNode)
     */
    public void describe (DNode n) {
        n.app("[").app(id_).app("] ");
        n.appShortClass(this).app(": ");
        n.app("atServer", isAtServer_).comma();
        n.app("stateless", isStateless_);
        // the widget models
        WModel [] wModels = getWModels();
        for (int i=0; i<wModels.length; i++) {
            if (wModels[i] != null) {
                DNode son = new DNode(n);
                wModels[i].describe(son);
            }
        }
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString () {
        return DNode.toString(this);
    }

    /**
     * Shows structure in a SWT window.
     */
    public void display () {
        DNode.display (this);
    }

    /**
     * Make <tt>n</tt> random changes to each widget model of this page
     */
    public void randomlyChange (int n) {
       WModel [] wModels = getWModels();
        for (int i=0; i<wModels.length; i++) {
            for (int j=0; j<n; j++) wModels[i].randomlyChange();
        }
    }

    /**
     * This debugging purpose method compares all widget models of
     * a client and server page if their state match at two
     * specified syncpoints.
     *
     * @param syncPoint 0 after a client to server sync, 1 after
     *         a server to client sync.
     *
     * @see WModel#equalsCS
     */
    public void equalsCS (Page mServer, int syncPoint) {
        WModel [] wModelsClient = getWModels();
        WModel [] wModelsServer = mServer.getWModels();
        if(wModelsClient.length!=wModelsServer.length) throw new RuntimeException();
        for (int i=0; i<wModelsClient.length; i++) {
            wModelsClient[i].equalsCS(wModelsServer[i], syncPoint);
        }
    }


    /**
     * Identifies the page within its component.
     */
    private short               id_;

    /**
     * Indicates that this Page executes at the server side.
     */
    private boolean             isAtServer_;

    /**
     * Specifies whether this page is stateless or not (at the server side).
     */
    private boolean             isStateless_;

    /** collects new model events for synchronization of dynamically added widget models to the other side. */
    private List				newModelEvents_;

    /** collects handled newModelEvents for sending in full synchronizations */
    private List				handledNewModelEvents_;

}
